# -*- coding: utf-8 -*-
"""
Created on Mon Jan  2 11:08:12 2023

@author: michael pawelzik
@co-author: uros mavric
Some parts have been copied from uros mavric and then they have been modified for compatibility. 

#### SYNTAX FOR CALLING CLASS OBJECTS ########################################

Look into module play_with_ahlborn.py

##############################################################################
"""



import time
from telnetlib import Telnet
import re
from datetime import datetime

class ahlborn:
   
    
    # constructor of class
    def __init__(self, ip = '192.168.115.44', wait_time = 1):
        self.__ip_adress = ip 
        self.__buffer = ''
        self.timestamp = datetime(1970,1,1,0,0,0)
        self.__tn = Telnet(self.__ip_adress, 10001)
        self.__wait_time = wait_time
        
        self.temp_entries = []
        
        # assign properties for temperature
        self.temp_ids = []
        self.temp_vals = []
        self.temp_units = []
        self.temp_val_categories = []
        self.temp_info_fields = []
        
        # assign properties for humidity
        self.hum_ids =[]
        self.hum_vals = []
        self.hum_units = []
        self.hum_val_categories = []
        self.hum_info_fields = []
        
        # assign properties for pressure
        self.press_ids =[]
        self.press_vals = []
        self.press_units = []
        self.press_val_categories = []
        self.press_info_fields = []
        
    # method that requests the sensor data from ahlborn
    def request_temp_hum_data_from_ahlborn(self):
        
        
        iteration = 0
        while iteration < 3:
            try:
                
             
                # timestamp for sensor data from ahlborn
                self.timestamp = datetime.now()
                
                # A command for manual request of sensor data is sent to ahlborn
                self.__tn.write(bytes('S1',encoding='utf8') + b'\r')
                time.sleep(self.__wait_time)
                
                # request timestamp from OS for sensor data
                self.__timestamp = datetime.now()
                
                # buffer of ahlborn is read and stored in buffer
                buffer = self.__tn.read_very_eager()
                time.sleep(self.__wait_time)
                
                # data of buffer is decoded with cp437
                self.__buffer = buffer.decode('cp437')
                time.sleep(self.__wait_time)
                
                if self.__buffer == '':
                    iteration += 1
                    print('An error has occured during readout of sensor data.')
                else:
                    
                    # leave while loop
                    break
            
            except EOFError:
                iteration += 1
                print('An error has occured during readout of sensor data.')
        
        return
       
    
    # method for extraction of sensor temperatures out of buffer of ahlborn    
    def get_param_of_temp_entries(self):
        
        self.temp_entries.clear()
        
        #clear properties for temperature
        self.temp_ids.clear()
        self.temp_vals.clear()
        self.temp_units.clear()
        self.temp_val_categories.clear()
        self.temp_info_fields.clear()
        
        # find all temperature entries in buffer of ahlborn
        list_temp_strings = re.findall(r'([0-9]{2}:.[+-].*\s[°C|K]+\s.*\r\n)', self.__buffer)
        
        # for each temperature entry split string into list of parameters. Delimiter is white space char 
        list_temp_entries = [re.split(r'\s', element, maxsplit=4) for element in list_temp_strings]
        
        
        # for each temp entry assign parameter to class property of parameter             
        for element in list_temp_entries:
           
            self.temp_ids.append(int(element[0].rstrip(':')))
            self.temp_vals.append(float(element[1].rstrip()))
            self.temp_units.append(element[2].rstrip())
            self.temp_val_categories.append(element[3].rstrip())
            self.temp_info_fields.append(element[4].rstrip())            
           
        return
        

    # method for extraction of sensor humidities out of buffer of ahlborn    
    def get_param_of_hum_entries(self):
                
        # clear properties for humidity
        self.hum_ids.clear()
        self.hum_vals.clear()
        self.hum_units.clear()
        self.hum_val_categories.clear()
        self.hum_info_fields.clear()
        
        # find all humidity entries in buffer of ahlborn
        list_hum_strings = re.findall(r'([0-9]{2}:.[+-].*\s%H\s.*\r\n)', self.__buffer)
        
        # for each humidity entry split string into list of parameters. Delimiter is white space char 
        list_hum_entries = [re.split(r'\s', element, maxsplit=4) for element in list_hum_strings]
        
        
        # for each hum entry assign parameter to class property of parameter             
        for element in list_hum_entries:
           
            self.hum_ids.append(int(element[0].rstrip(':')))
            self.hum_vals.append(float(element[1].rstrip()))
            self.hum_units.append(element[2].rstrip())
            self.hum_val_categories.append(element[3].rstrip())
            self.hum_info_fields.append(element[4].rstrip())
        
        return

    def get_param_of_press_entries(self):
        
                
        # clear properties for pressure
        self.press_ids.clear()
        self.press_vals.clear()
        self.press_units.clear()
        self.press_val_categories.clear()
        self.press_info_fields.clear()
        
        # find all pressure entries in buffer of ahlborn
        list_press_strings = re.findall(r'([0-9]{2}:.[+-].*\smb\s.*\r\n)', self.__buffer)
        
        # for each pressure entry split string into list of parameters. Delimiter is white space char 
        list_press_entries = [re.split(r'\s', element, maxsplit=4) for element in list_press_strings]
        
        
        # for each pressure entry assign parameter to class property of parameter             
        for element in list_press_entries:
           
            self.press_ids.append(int(element[0].rstrip(':')))
            self.press_vals.append(float(element[1].rstrip()))
            self.press_units.append(element[2].rstrip())
            self.press_val_categories.append(element[3].rstrip())
            self.press_info_fields.append(element[4].rstrip()) 
           
        return
    
    # close telenet connection
    def close(self):
        self.__tn.close()
    
    # destructor of class
    def __del__(self):
        self.close()



    
    